/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Matrix.h"
#include "Map.h"

using namespace std;
using namespace FD;

class ResizeMapMatrixValues;

DECLARE_NODE(ResizeMapMatrixValues)
/*Node
 *
 * @name ResizeMapMatrixValues
 * @category HARK:MISC
 * @description Resize Matrix value in Map<int,ObjectRef>. ObjectRef will be Matrix<int>, Matrix<float> or Matrix<complex<float> >.
 *
 * @input_name INPUT
 * @input_type Map<int,ObjectRef>
 * @input_description The input Map. Map<int,Matrix<int> >, Map<int,Matrix<float> > or Map<int,Matrix<complex<float> > >.
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description The Map with the resized value. Map<int,Matrix<int> >, Map<int,Matrix<float> > or Map<int,Matrix<complex<float> > >.
 *
 * @parameter_name RESIZE_TYPE
 * @parameter_type string
 * @parameter_value RELATIVE
 * @parameter_list RELATIVE:ABSOLUTE
 * @parameter_description Resize type. Suppose current vector size is (A,B). If RELATIVE, the vector size would be (A+SIZE_ROW,B+SIZE_COL). If ABSOLUTE, the vector size would be (SIZE_ROW,SIZE_COL). [default: RELATIVE] 
 *
 * @parameter_name SIZE_ROW
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description Number of added row elements. If negative, the elements will be deleted from the end. If positive, zero elements are added at the end. [default: 0]
 *
 * @parameter_name SIZE_COLUMN
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description Number of added column elements. If negative, the elements will be deleted from the end. If positive, zero elements are added at the end. [default: 0]
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Show debug points. [default: false]
 *
 END*/

static const int SIZE_RELATIVE = 0;
static const int SIZE_ABSOLUTE = 1;

class ResizeMapMatrixValues : public BufferedNode {
	int inputID;
	int outputID;
	int resize_type;  
	int size_row;
	int size_column;
	bool debug;

public:
	ResizeMapMatrixValues(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		inputID = addInput("INPUT");
		outputID = addOutput("OUTPUT");
		string type_str = object_cast<String>(parameters.get("RESIZE_TYPE"));
		if (type_str == "RELATIVE")      resize_type = SIZE_RELATIVE;
		else if (type_str == "ABSOLUTE") resize_type = SIZE_ABSOLUTE;
		else {
			cerr << "******* ResizeMapMatrixValues: Input error" << endl;
			throw new NodeException(NULL, "ResizeMapMatrixValues: SIZE_TYPE should be \"RELATIVE\" or \"ABSOLUTE\".", __FILE__, __LINE__);
		}
		size_row = dereference_cast<int>(parameters.get("SIZE_ROW"));
		size_column = dereference_cast<int>(parameters.get("SIZE_COLUMN"));
		debug = dereference_cast<bool>(parameters.get("DEBUG"));
	}

	int checkSize(int vSize, int size) {
		int fsize = 0;
		if (resize_type == SIZE_RELATIVE)      fsize = vSize + size;
		else if (resize_type == SIZE_ABSOLUTE) fsize = size;
		if (fsize < 0) {
			cerr << "******* ResizeMapMatrixValues: Resized size should not be negative." << endl;
			throw(new NodeException(NULL, "Exception caught in calculate[ResizeMapMatrixValues]", __FILE__, __LINE__));
		}
		return fsize;
	}

	void calculate(int output_id, int count, Buffer &out) {
 		RCPtr<Map<int, ObjectRef> > input_ptr =  getInput(inputID, count);
		const Map<int, ObjectRef> input_map = *input_ptr;
     
		if (debug == true) {
			cout << "******* ResizeMapMatrixValues debug point-Input check (count=" << count << ")" << endl;
 			if (input_map.empty() == true) {
				cout << "map empty" << endl;
			}
			else {
				cout << "Input map begins from " << (input_map.begin()->first) << endl;
				cout << "Number of channels: " << input_map.size() << endl;
				cout << "Type of first channel of ObjectRef of input map is: ";
				if (typeid(*(input_map.begin()->second))      == typeid(Matrix<int>))			  cout << "Matrix<int>" << endl;
				else if (typeid(*(input_map.begin()->second)) == typeid(Matrix<float>))           cout << "Matrix<float>" << endl;
				else if (typeid(*(input_map.begin()->second)) == typeid(Matrix<complex<float> >)) cout << "Matrix<complex<float> >" << endl;
				else                                                                              cout << "unidentified." << endl;
			}
		}

		Map<int, ObjectRef>& output = *new Map<int, ObjectRef>;
		out[count] = &output;

		if (input_map.empty() == false) {
			Map<int, ObjectRef>::const_iterator it = input_map.begin();

			if (typeid(*(it->second)) == typeid(Matrix<int>)) {
				for (; it != input_map.end(); ++it) {
					const Matrix<int> &vcInput = object_cast<Matrix<int> >(it->second);
					size_t rSize = vcInput.nrows();
					size_t cSize = vcInput.ncols();
					int fRsize = checkSize(rSize, size_row);
					int fCsize = checkSize(cSize, size_column);
					RCPtr<Matrix<int> > fbank(new Matrix<int>(fRsize, fCsize));
					for (int i = 0; i < fRsize; ++i) {
						if (i >= rSize) break;
						for (int j = 0; j < fCsize; ++j) {
							if (j >= cSize) break;
							(*fbank)[i][j] = vcInput[i][j];
						}
					}
					output[it->first] = fbank;
				}
			}
			else if (typeid(*(it->second)) == typeid(Matrix<float>)) {
				for (; it != input_map.end(); ++it) {
					const Matrix<float> &vcInput = object_cast<Matrix<float> >(it->second);
					size_t rSize = vcInput.nrows();
					size_t cSize = vcInput.ncols();
					int fRsize = checkSize(rSize, size_row);
					int fCsize = checkSize(cSize, size_column);
					RCPtr<Matrix<float> > fbank(new Matrix<float>(fRsize, fCsize));
					for (int i = 0; i < fRsize; ++i) {
						if (i >= rSize) break;
						for (int j = 0; j < fCsize; ++j) {
							if (j >= cSize) break;
							(*fbank)[i][j] = vcInput[i][j];
						}
					}
					output[it->first] = fbank;
				}
			}
			else if (typeid(*(it->second)) == typeid(Matrix<complex<float> >)) {
				for (; it != input_map.end(); ++it) {
					const Matrix<complex<float> > &vcInput = object_cast<Matrix<complex<float> > >(it->second);
					size_t rSize = vcInput.nrows();
					size_t cSize = vcInput.ncols();
					int fRsize = checkSize(rSize, size_row);
					int fCsize = checkSize(cSize, size_column);
					RCPtr<Matrix<complex<float> > > fbank(new Matrix<complex<float> >(fRsize, fCsize));
					for (int i = 0; i < fRsize; ++i) {
						if (i >= rSize) break;
						for (int j = 0; j < fCsize; ++j) {
							if (j >= cSize) break;
							(*fbank)[i][j] = vcInput[i][j];
						}
					}
					output[it->first] = fbank;
				}
			}
			else {
				cerr << "******* ResizeMapMatrixValues: Input type is unidentified." << endl;
				throw(new NodeException(NULL, "Exception caught in calculate[ResizeMapMatrixValues]", __FILE__, __LINE__));
			}
		}		

        if (debug == true) cout << "******* ResizeMapMatrixValues debug point-Finish" << endl;
	}

	IN_ORDER_NODE_SPEEDUP(ResizeMapMatrixValues);
};
