/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include <math.h>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"
#include "Source.h"

using namespace std;
using namespace FD;

class SourceSelectorBySourceInfo;

DECLARE_NODE(SourceSelectorBySourceInfo);
/*Node
 *
 * @name SourceSelectorBySourceInfo
 * @category HARK:MISC
 * @description Select sources within the specified source information.
 *
 * @input_name SOURCES
 * @input_type Vector<ObjectRef>
 * @input_description Source locations with ID. Each element of the vector is a source location with ID specified by "Source".
 *
 * @output_name OUTPUT
 * @output_type Vector<ObjectRef>
 * @output_description Source locations with ID within the specified source information.
 *
 * @parameter_name SELECTION_TYPE
 * @parameter_type string
 * @parameter_value ID
 * @parameter_list ALL:ID:POWER:DIRECTION_AZIMUTH:DIRECTION_ELEVATION
 * @parameter_description Criterion type for selecting Source. If SOURCES is not connected, ID should be selected.
 * 
 * @parameter_name ID_SELECTION_TYPE
 * @parameter_type string
 * @parameter_value LATEST
 * @parameter_list LATEST:OLDEST:BETWEEN
 * @parameter_valid SELECTION_TYPE=ID
 * @parameter_description Criterion for selecting Source using ID.
 * 
 * @parameter_name ID_RANGE_MIN
 * @parameter_type int
 * @parameter_value 0
 * @parameter_valid ID_SELECTION_TYPE=BETWEEN
 * @parameter_description Minimum ID for source selection.
 * 
 * @parameter_name ID_RANGE_MAX
 * @parameter_type int
 * @parameter_value 0
 * @parameter_valid ID_SELECTION_TYPE=BETWEEN
 * @parameter_description Maximum ID for source selection.
 * 
 * @parameter_name POWER_SELECTION_TYPE
 * @parameter_type string
 * @parameter_value HIGHEST
 * @parameter_list HIGHEST:LOWEST:BETWEEN
 * @parameter_valid SELECTION_TYPE=POWER
 * @parameter_description Criterion for selecting Source using POWER.
 * 
 * @parameter_name POWER_RANGE_MIN
 * @parameter_type float
 * @parameter_value 0
 * @parameter_valid POWER_SELECTION_TYPE=BETWEEN
 * @parameter_description Minimum POWER for source selection.
 * 
 * @parameter_name POWER_RANGE_MAX
 * @parameter_type float
 * @parameter_value 40.0
 * @parameter_valid POWER_SELECTION_TYPE=BETWEEN
 * @parameter_description Maximum POWER for source selection.
 * 
 * @parameter_name AZIMUTH_SELECTION_TYPE
 * @parameter_type string
 * @parameter_value BETWEEN
 * @parameter_list NEAREST:BETWEEN
 * @parameter_valid SELECTION_TYPE=DIRECTION_AZIMUTH
 * @parameter_description Criterion for selecting Source using DIRECTION_AZIMUTH.
 * 
 * @parameter_name AZIMUTH
 * @parameter_type float
 * @parameter_value 0.0
 * @parameter_valid AZIMUTH_SELECTION_TYPE=NEAREST
 * @parameter_description Azimuth to select Source when NEAREST is selected.
  * 
 * @parameter_name AZIMUTH_RANGE_MIN
 * @parameter_type float
 * @parameter_value 0.0
 * @parameter_valid AZIMUTH_SELECTION_TYPE=BETWEEN
 * @parameter_description Minimum Azimuth for source selection.
 * 
 * @parameter_name AZIMUTH_RANGE_MAX
 * @parameter_type float
 * @parameter_value 360.0
 * @parameter_valid AZIMUTH_SELECTION_TYPE=BETWEEN
 * @parameter_description Maximum Azimuth for source selection.
 *
 * @parameter_name ELEVATION_SELECTION_TYPE
 * @parameter_type string
 * @parameter_value BETWEEN
 * @parameter_list NEAREST:BETWEEN
 * @parameter_valid SELECTION_TYPE=DIRECTION_ELEVATION
 * @parameter_description Criterion for selecting Source using DIRECTION_ELEVATION.
 * 
 * @parameter_name ELEVATION
 * @parameter_type float
 * @parameter_value 0.0
 * @parameter_valid ELEVATION_SELECTION_TYPE=NEAREST
 * @parameter_description Elevation to select Source when NEAREST is selected.
  * 
 * @parameter_name ELEVATION_RANGE_MIN
 * @parameter_type float
 * @parameter_value -90.0
 * @parameter_valid ELEVATION_SELECTION_TYPE=BETWEEN
 * @parameter_description Minimum Elevation for source selection.
 * 
 * @parameter_name ELEVATION_RANGE_MAX
 * @parameter_type float
 * @parameter_value 90.0
 * @parameter_valid ELEVATION_SELECTION_TYPE=BETWEEN
 * @parameter_description Maximum Elevation for source selection.
 *
 END*/

static const int SELECTION_ALL = 0;
static const int SELECTION_ID = 1;
static const int SELECTION_POWER = 2;
static const int SELECTION_AZIMUTH = 3;
static const int SELECTION_ELEVATION = 4;

static const int ID_SELECTION_LATEST = 0;
static const int ID_SELECTION_OLDEST = 1;
static const int ID_SELECTION_BETWEEN = 2;

static const int POWER_SELECTION_HIGHEST = 0;
static const int POWER_SELECTION_LOWEST = 1;
static const int POWER_SELECTION_BETWEEN = 2;

static const int DIRECTION_SELECTION_NEAREST = 0;
static const int DIRECTION_SELECTION_BETWEEN = 1;

class SourceSelectorBySourceInfo : public BufferedNode {
	int sourcesID;
	int outputID;

	int selection_type;

	int id_selection_type;
	int max_id;
	int min_id;

	int power_selection_type;
	float max_power;
	float min_power;

	int azimuth_selection_type;
	float azimuth;
	float max_azimuth;
	float min_azimuth;

	int elevation_selection_type;
	float elevation;
	float max_elevation;
	float min_elevation;

	//bool debug;

public:
	SourceSelectorBySourceInfo(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		string type_str;

		sourcesID = addInput("SOURCES");
		outputID = addOutput("OUTPUT");

		type_str = object_cast<string>(parameters.get("SELECTION_TYPE"));
		if (type_str == "ALL")                      selection_type = SELECTION_ALL;
		else if (type_str == "ID")                  selection_type = SELECTION_ID;
		else if (type_str == "POWER")               selection_type = SELECTION_POWER;
		else if (type_str == "DIRECTION_AZIMUTH")   selection_type = SELECTION_AZIMUTH;
		else if (type_str == "DIRECTION_ELEVATION") selection_type = SELECTION_ELEVATION;
		else {
			cerr << "******* SourceSelectorBySourceInfo: Input error" << endl;
			throw new NodeException(NULL, "SourceSelectorBySourceInfo: SELECTION_TYPE should be \"ALL\" or \"ID\" or \"POWER\" or \"DIRECTION_AZIMUTH\" or \"DIRECTION_ELEVATION\".", __FILE__, __LINE__);
		}

		if (selection_type == SELECTION_ID) {
			type_str = object_cast<string>(parameters.get("ID_SELECTION_TYPE"));
			if (type_str == "LATEST")       id_selection_type = ID_SELECTION_LATEST;
			else if (type_str == "OLDEST")  id_selection_type = ID_SELECTION_OLDEST;
			else if (type_str == "BETWEEN") id_selection_type = ID_SELECTION_BETWEEN;
			else {
				cerr << "******* SourceSelectorBySourceInfo: Input error" << endl;
				throw new NodeException(NULL, "SourceSelectorBySourceInfo: ID_SELECTION_TYPE should be \"LATEST\" or \"OLDEST\" or \"BETWEEN\".", __FILE__, __LINE__);
			}
			if (id_selection_type == ID_SELECTION_BETWEEN) {
				min_id = dereference_cast<int>(parameters.get("ID_RANGE_MIN"));     
				max_id = dereference_cast<int>(parameters.get("ID_RANGE_MAX"));   
				if (max_id < min_id) {
					cerr << "******* SourceSelectorBySourceInfo: Input error" << endl;
					throw new NodeException(NULL, string("SourceSelectorBySourceInfo: max_id must not be smaller than min_id."), __FILE__, __LINE__);
				}
				if (min_id < 0 || max_id < 0) {
					cerr << "******* SourceSelectorBySourceInfo: Input error" << endl;
					throw new NodeException(NULL, string("SourceSelectorBySourceInfo: ID_RANGE must be lager than or equal 0."), __FILE__, __LINE__);
				}
			}
		}

		if (selection_type == SELECTION_POWER) {
			type_str = object_cast<string>(parameters.get("POWER_SELECTION_TYPE"));
			if (type_str == "HIGHEST")      power_selection_type = POWER_SELECTION_HIGHEST;
			else if (type_str == "LOWEST")  power_selection_type = POWER_SELECTION_LOWEST;
			else if (type_str == "BETWEEN") power_selection_type = POWER_SELECTION_BETWEEN;
			else {
				cerr << "******* SourceSelectorBySourceInfo: Input error" << endl;
				throw new NodeException(NULL, "SourceSelectorBySourceInfo: POWER_SELECTION_TYPE should be \"HIGHEST\" or \"LOWEST\" or \"BETWEEN\".", __FILE__, __LINE__);
			}
			if (power_selection_type == POWER_SELECTION_BETWEEN) {
				min_power = dereference_cast<float>(parameters.get("POWER_RANGE_MIN"));        
				max_power = dereference_cast<float>(parameters.get("POWER_RANGE_MAX"));
				if (max_power < min_power){
					cerr << "******* SourceSelectorBySourceInfo: Input error" << endl;
					throw new NodeException(NULL, string("SourceSelectorBySourceInfo: POWER_RANGE is not appropriate."), __FILE__, __LINE__);
				}
			}
		}

		if (selection_type == SELECTION_AZIMUTH) {
			type_str = object_cast<string>(parameters.get("AZIMUTH_SELECTION_TYPE"));
			if (type_str == "NEAREST")      azimuth_selection_type = DIRECTION_SELECTION_NEAREST;
			else if (type_str == "BETWEEN") azimuth_selection_type = DIRECTION_SELECTION_BETWEEN;
			else {
				cerr << "******* SourceSelectorBySourceInfo: Input error" << endl;
				throw new NodeException(NULL, "SourceSelectorBySourceInfo: AZIMUTH_SELECTION_TYPE should be \"NEAREST\" or \"BETWEEN\".", __FILE__, __LINE__);
			}
			if (azimuth_selection_type == DIRECTION_SELECTION_NEAREST) {
				azimuth = dereference_cast<float>(parameters.get("AZIMUTH"));
				while (azimuth < -180) azimuth += 360;
				while (azimuth > 180)  azimuth -= 360;
			}
			else if (azimuth_selection_type == DIRECTION_SELECTION_BETWEEN) {
				min_azimuth = dereference_cast<float>(parameters.get("AZIMUTH_RANGE_MIN"));   
				max_azimuth = dereference_cast<float>(parameters.get("AZIMUTH_RANGE_MAX"));     
				if (max_azimuth < min_azimuth) {
					cerr << "******* SourceSelectorBySourceInfo: Input error" << endl;
			  		throw new NodeException(NULL, string("AZIMUTH_RANGE is not appropriate."), __FILE__, __LINE__);
				} 	
				while (min_azimuth < -180) min_azimuth += 360;
				while (min_azimuth > 180)  min_azimuth -= 360;
				while (max_azimuth < -180) max_azimuth += 360;
				while (max_azimuth > 180)  max_azimuth -= 360;
			}
		}

		if (selection_type == SELECTION_ELEVATION) {
			type_str = object_cast<string>(parameters.get("ELEVATION_SELECTION_TYPE"));
			if (type_str == "NEAREST")      elevation_selection_type = DIRECTION_SELECTION_NEAREST;
			else if (type_str == "BETWEEN") elevation_selection_type = DIRECTION_SELECTION_BETWEEN;
			else {
				cerr << "******* SourceSelectorBySourceInfo: Input error" << endl;
				throw new NodeException(NULL, "SourceSelectorBySourceInfo: ELEVATION_SELECTION_TYPE should be \"NEAREST\" or \"BETWEEN\".", __FILE__, __LINE__);
			}
			if (elevation_selection_type == DIRECTION_SELECTION_NEAREST) {
				elevation = dereference_cast<float>(parameters.get("ELEVATION"));
				while (elevation < -180) elevation += 360;
				while (elevation > 180)  elevation -= 360;
			}
			else if (elevation_selection_type == DIRECTION_SELECTION_BETWEEN) {
				min_elevation = dereference_cast<float>(parameters.get("ELEVATION_RANGE_MIN"));   
				max_elevation = dereference_cast<float>(parameters.get("ELEVATION_RANGE_MAX"));     
				if (max_elevation < min_elevation) {
					cerr << "******* SourceSelectorBySourceInfo: Input error" << endl;
			  		throw new NodeException(NULL, string("ELEVATION_RANGE is not appropriate."), __FILE__, __LINE__);
				} 	
				while (min_elevation < -180) min_elevation += 360;
				while (min_elevation > 180)  min_elevation -= 360;
				while (max_elevation < -180) max_elevation += 360;
				while (max_elevation > 180)  max_elevation -= 360;
			}
		}

		//debug = dereference_cast<bool>(parameters.get("DEBUG"));

    	inOrder = true;
	}

	void calculate(int output_id, int count, Buffer &out) {
		RCPtr<Vector<ObjectRef> > sources = getInput(sourcesID, count);

		//if (debug == true) cout << "******* SourceSelectorBySourceInfo debug point (count=" << count << ")" << endl;

		RCPtr<Vector<ObjectRef> > selected_sources(new Vector<ObjectRef>);
		out[count] = selected_sources;

		RCPtr<Source> src;
		int id;
		switch (selection_type) {
			case SELECTION_ALL:
				for (int i = 0; i < sources->size(); i++) {
					src = (*sources)[i];
					selected_sources->push_back(src);
				}
				break;					
			case SELECTION_ID:
				if (id_selection_type == ID_SELECTION_LATEST || id_selection_type == ID_SELECTION_OLDEST) {
					src = (*sources)[0];
					id = src->id;
					for (int i = 1; i < sources->size(); i++) {
						src = (*sources)[i];
						if ((id_selection_type == ID_SELECTION_LATEST && src->id > id) || 
                            (id_selection_type == ID_SELECTION_OLDEST && src->id < id)) {
							id = src->id;
						}
					}
					src = (*sources)[id];
					selected_sources->push_back(src);
				}
				else { //ID_SELECTION_BETWEEN
					for (int i = 0; i < sources->size(); i++) {
						src = (*sources)[i];
						if (src->id >= min_id && src->id <= max_id) {
							selected_sources->push_back(src);
						}
					}
				}
				break;
			case SELECTION_POWER:
				if (power_selection_type == POWER_SELECTION_HIGHEST || power_selection_type == POWER_SELECTION_LOWEST) {
					src = (*sources)[0];
					id = src->id;
					float power = src->power;
					for (int i = 1; i < sources->size(); i++) {
						src = (*sources)[i];
						if ((power_selection_type == POWER_SELECTION_HIGHEST && src->power > power) || 
                            (power_selection_type == POWER_SELECTION_LOWEST  && src->power < power)) {
							power = src->power;
							id = src->id;
						}
					}
					src = (*sources)[id];
					selected_sources->push_back(src);
				}
				else { //POWER_SELECTION_BETWEEN
					for (int i = 0; i < sources->size(); i++) {
						src = (*sources)[i];
						if (src->power >= min_power && src->power <= max_power) {
							selected_sources->push_back(src);
						}
					}
				}
				break;
			case SELECTION_AZIMUTH:
				float adir;
				if (azimuth_selection_type == DIRECTION_SELECTION_NEAREST) {
					src = (*sources)[0];
					id = src->id;
					float dir = 180.0f / M_PI * atan2(src->x[1], src->x[0]);
					for (int i = 1; i < sources->size(); i++) {
						src = (*sources)[i];
						adir = 180.0f / M_PI * atan2(src->x[1], src->x[0]);
	    				if (abs(adir - azimuth) < abs(dir - azimuth)) {
							dir = adir;
							id = src->id;
						}
					}
					src = (*sources)[id];
					selected_sources->push_back(src);
				}
				else { //DIRECTION_SELECTION_BETWEEN
					for (int i = 0; i < sources->size(); i++) {
						src = (*sources)[i];
						adir = 180.0f / M_PI * atan2(src->x[1], src->x[0]);
						if (adir >= min_azimuth && adir <= max_azimuth) {
							selected_sources->push_back(src);
						}
					}
				}
				break;
			case SELECTION_ELEVATION:
				float edir;
				if (elevation_selection_type == DIRECTION_SELECTION_NEAREST) {
					src = (*sources)[0];
					id = src->id;
					float dir = 180.0f / M_PI * atan2(src->x[2], sqrt(src->x[1] * src->x[1] + src->x[0] * src->x[0]));
					for (int i = 1; i < sources->size(); i++) {
						src = (*sources)[i];
						edir = 180.0f / M_PI * atan2(src->x[2], sqrt(src->x[1] * src->x[1] + src->x[0] * src->x[0]));
	    				if (abs(edir - elevation) < abs(dir - elevation)) {
							dir = edir;
							id = src->id;
						}
					}
					src = (*sources)[id];
					selected_sources->push_back(src);
				}
				else { //DIRECTION_SELECTION_BETWEEN
					for (int i = 0; i < sources->size(); i++) {
						src = (*sources)[i];
						edir = 180.0f / M_PI * atan2(src->x[2], sqrt(src->x[1] * src->x[1] + src->x[0] * src->x[0]));
						if (edir >= min_elevation && edir <= max_elevation) {
							selected_sources->push_back(src);
						}
					}
				}
				break;
			default:
				break;
		}

        //if (debug == true) cout << "******* SourceSelectorBySourceInfo debug point-Finish" << endl;
	}

	IN_ORDER_NODE_SPEEDUP(SourceSelectorBySourceInfo);
};
