/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"

using namespace FD;
using namespace std;

class VectorValueOverwrite;

DECLARE_NODE(VectorValueOverwrite)
/*Node
 *
 * @name VectorValueOverwrite
 * @category HARK:MISC
 * @description Replaces specific elements of the Vector by a given value.
 *
 * @input_name INPUT
 * @input_type any
 * @input_description The input Vector. Vector<int>, Vector<float> or Vector<complex<float> >.
 *
 * @output_name OUTPUT
 * @output_type any
 * @output_description The Vector with the overwritten value. Vector<int>, Vector<float> or Vector<complex<float> >.
 *
 * @parameter_name OVERWRITTEN_MIN
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description The index where to start setting the Vector.
 *
 * @parameter_name OVERWRITTEN_MAX
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description The index where to end setting the Vector.
 *
 * @parameter_name OVERWRITE_VALUE_REAL
 * @parameter_type float
 * @parameter_value 0
 * @parameter_description Real part of setting value.
 *
 * @parameter_name OVERWRITE_VALUE_IMAG
 * @parameter_type float
 * @parameter_value 0
 * @parameter_description Imaginary part of setting value.
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Show debug points. [default: false]
 *
 END*/

class VectorValueOverwrite : public BufferedNode {
	int inputID;
	int outputID;
	int vMin;
	int vMax;
	float fReal;
	float fImag;	
	bool debug;

public:
	VectorValueOverwrite(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		inputID = addInput("INPUT");
		outputID = addOutput("OUTPUT");
		vMin = dereference_cast<int>(parameters.get("OVERWRITTEN_MIN"));
		vMax = dereference_cast<int>(parameters.get("OVERWRITTEN_MAX"));
		fReal = dereference_cast<float>(parameters.get("OVERWRITE_VALUE_REAL"));
		fImag = dereference_cast<float>(parameters.get("OVERWRITE_VALUE_IMAG"));
		debug = dereference_cast<bool>(parameters.get("DEBUG"));
	}

	void checkSize(int &vMin, int &vMax, int vSize) {
		if (vMin < 0) vMin = 0;
		else if (vMin >= vSize) vMin = vSize - 1;
		if (vMax < 0) vMax = 0;
		else if (vMax >= vSize) vMax = vSize - 1;
		if (vMin > vMax) {
			cerr << "******* VectorValueOverwrite: Minimun index is larger than maximum index." << endl;
			throw(new NodeException(NULL, "Exception caught in calculate[VectorValueOverwrite]", __FILE__, __LINE__));
		}
	}

	void calculate(int output_id, int count, Buffer &out) {
		ObjectRef input = getInput(inputID, count);

		if (debug == true) {
			cout << "******* VectorValueOverwrite debug point-Input check (count=" << count << ")" << endl;
            cout << "Input type is: ";
			if (typeid(*input) == typeid(Vector<int>))					cout << "Vector<int>" << endl;
			else if (typeid(*input) == typeid(Vector<float>))           cout << "Vector<float>" << endl;
			else if (typeid(*input) == typeid(Vector<complex<float> >)) cout << "Vector<complex<float> >" << endl;
			else                                                        cout << "unidentified." << endl;
		}

		if (typeid(*input) == typeid(Vector<int>)) {
			RCPtr<Vector<int> > inputPtr = getInput(inputID, count);
			size_t vSize = inputPtr->size();
			checkSize(vMin, vMax, vSize);
			//RCPtr<Vector<int> > output(Vector<int>::alloc(vSize));
			RCPtr<Vector<int> > output(new Vector<int>(vSize));
			/*output = *inputPtr;
			for (int i = vMin; i <= vMax; i++) {
				(*output)[i] = static_cast<int>(fReal);
			}*/
			int i = 0;
			for (; i < vMin; i++)  (*output)[i] = (*inputPtr)[i];
			for (; i <= vMax; i++) (*output)[i] = static_cast<int>(fReal);
			for (; i < vSize; i++) (*output)[i] = (*inputPtr)[i];
			out[count] = output;
		}
		else if (typeid(*input) == typeid(Vector<float>)) {
			RCPtr<Vector<float> > inputPtr = getInput(inputID, count);
			size_t vSize = inputPtr->size();
			checkSize(vMin, vMax, vSize);
			//RCPtr<Vector<float> > output(Vector<float>::alloc(vSize));
			RCPtr<Vector<float> > output(new Vector<float>(vSize));
			/**output = *inputPtr;
			for (int i = vMin; i <= vMax; i++) {
				(*output)[i] = fReal;
			}*/
			int i = 0;
			for (; i < vMin; i++)  (*output)[i] = (*inputPtr)[i];
			for (; i <= vMax; i++) (*output)[i] = fReal;
			for (; i < vSize; i++) (*output)[i] = (*inputPtr)[i];
			out[count] = output;
		}
		else if (typeid(*input) == typeid(Vector<complex<float> >)) {
			RCPtr<Vector<complex<float> > > inputPtr = getInput(inputID, count);
			size_t vSize = inputPtr->size();
			checkSize(vMin, vMax, vSize);
			//RCPtr<Vector<complex<float> > > output(Vector<complex<float> >::alloc(vSize));
			RCPtr<Vector<complex<float> > > output(new Vector<complex<float> >(vSize));
			/**output = *inputPtr;
			for (int i = vMin; i <= vMax; i++) {
				(*output)[i] = complex<float>(fReal, fImag);
			}*/
			int i = 0;
			for (; i < vMin; i++)  (*output)[i] = (*inputPtr)[i];
			for (; i <= vMax; i++) (*output)[i] = complex<float>(fReal, fImag);
			for (; i < vSize; i++) (*output)[i] = (*inputPtr)[i];
			out[count] = output;
		}
		else {
			cerr << "******* VectorValueOverwrite: Input type is unidentified." << endl;
			throw(new NodeException(NULL, "Exception caught in calculate[VectorValueOverwrite]", __FILE__, __LINE__));
		}

        if (debug == true) cout << "******* VectorValueOverwrite debug point-Finish" << endl;
	}

    NO_ORDER_NODE_SPEEDUP(VectorValueOverwrite);
};
